import unittest
import numpy as np
from deepchem.utils import hash_utils


def random_string(length, chars=None):
  import string
  if chars is None:
    chars = list(string.ascii_letters + string.ascii_letters + '()[]+-.=#@/\\')
  return ''.join(np.random.choice(chars, length))


class TestHashUtils(unittest.TestCase):

  def test_hash_ecfp(self):
    for power in (2, 16, 64):
      for _ in range(10):
        string = random_string(10)
        string_hash = hash_utils.hash_ecfp(string, power)
        self.assertIsInstance(string_hash, int)
        self.assertLess(string_hash, 2**power)
        self.assertGreaterEqual(string_hash, 0)

  def test_hash_ecfp_pair(self):
    for power in (2, 16, 64):
      for _ in range(10):
        string1 = random_string(10)
        string2 = random_string(10)
        pair_hash = hash_utils.hash_ecfp_pair((string1, string2), power)
        self.assertIsInstance(pair_hash, int)
        self.assertLess(pair_hash, 2**power)
        self.assertGreaterEqual(pair_hash, 0)

  def test_vectorize(self):
    size = 16
    feature_dict = {0: "C", 1: "CC", 2: "CCC"}
    hash_function = hash_utils.hash_ecfp
    vector = hash_utils.vectorize(hash_function, feature_dict, size)
    assert vector.shape == (size,)
    assert np.count_nonzero(vector) == len(feature_dict)



import numpy as np
from deepchem.feat.base_classes import Featurizer
from deepchem.utils import get_partial_charge

from typing import Sequence


class AtomicConformation:
  """This class represents a collection of atoms arranged in 3D space.

  An instance of this class may represent any collection of atoms: a molecule,
  a fragment of a molecule, multiple interacting molecules, a material, etc.
  For each atom it stores a position and a list of scalar properties.  Arbitrary
  properties are supported, but convenience methods are provided for accessing
  certain standard ones: atomic number, formal charge, and partial charge.

  Instances of this class are most often created by AtomicConformationFeaturizer.

  Attributes
  ----------
  positions: ndarray
    the positions of all atoms in Angstroms, stored in an array of shape (N, 3)
    where N is the number of atoms
  properties: ndarray
    the property values for all atoms, stored in an array of shape (N, M) where
    N is the number of atoms and M is the number of properties
  property_names: ndarray
    an array of length M with the names of the properties
  """

  def __init__(self, positions: np.ndarray, properties: np.ndarray,
               property_names: Sequence[str]):
    """Create an AtomicConformation for a set of atoms.

    Parameters
    ----------
    positions: ndarray
      the positions of all atoms in Angstroms, stored in an array of shape (N, 3)
      where N is the number of atoms
    properties: ndarray
      the property values for all atoms, stored in an array of shape (N, M) where
      N is the number of atoms and M is the number of properties
    property_names: Sequence[str]
      the names of the properties
    """
    self.positions = positions
    self.properties = properties
    self.property_names = np.array(property_names)

  @property
  def num_atoms(self) -> int:
    """Get the number of atoms in this object."""
    return self.positions.shape[0]

  def get_property(self, name: str) -> np.ndarray:
    """Get a column of the properties array corresponding to a particular property.

    If there is no property with the specified name, this raises a ValueError.

    Parameters
    ----------
    name: str
      the name of the property to get

    Returns
    -------
    a numpy array containing the requested column of the properties array.  This
    is a 1D array of length num_atoms.
    """
    indices = np.where(self.property_names == name)[0]
    if len(indices) == 0:
      raise ValueError("No property called '%s'" % name)
    return self.properties[:, indices[0]]

  @property
  def atomic_number(self) -> np.ndarray:
    """Get the column of the properties array containing atomic numbers.

    If there is no property with the name 'atomic number', this raises a ValueError.

    Returns
    -------
    a numpy array containing the requested column of the properties array.  This
    is a 1D array of length num_atoms.
    """
    return self.get_property('atomic number')

  @property
  def formal_charge(self) -> np.ndarray:
    """Get the column of the properties array containing formal charges.

    If there is no property with the name 'formal charge', this raises a ValueError.

    Returns
    -------
    a numpy array containing the requested column of the properties array.  This
    is a 1D array of length num_atoms.
    """
    return self.get_property('formal charge')

  @property
  def partial_charge(self) -> np.ndarray:
    """Get the column of the properties array containing partial charges.

    If there is no property with the name 'partial charge', this raises a ValueError.

    Returns
    -------
    a numpy array containing the requested column of the properties array.  This
    is a 1D array of length num_atoms.
    """
    return self.get_property('partial charge')


class AtomicConformationFeaturizer(Featurizer):
  """This featurizer represents each sample as an AtomicConformation object,
  representing a 3D arrangement of atoms.

  It expects each datapoint to be a string, which may be either a filename or a
  SMILES string.  It is processed as follows.

  If the string ends in .pdb, .sdf, or .mol2, it is assumed to be a file in the
  corresponding format.  The positions and elements of all atoms contained in
  the file are loaded.  RDKit is used to compute formal and partial charges.

  Otherwise, it is assumed to be a SMILES string.  RDKit is used to generate a
  3D conformation and to compute formal and partial charges.

  Examples
  --------
  >>> import deepchem as dc
  >>> smiles = ['CCC']
  >>> featurizer = dc.feat.AtomicConformationFeaturizer()
  >>> features = featurizer.featurize(smiles)
  >>> features[0].num_atoms
  11
  >>> sum(features[0].atomic_number == 6)
  3
  >>> sum(features[0].atomic_number == 1)
  8
  >>> type(features[0].formal_charge)
  <class 'numpy.ndarray'>
  >>> features[0].formal_charge.shape
  (11,)
  >>> type(features[0].partial_charge)
  <class 'numpy.ndarray'>
  >>> features[0].partial_charge.shape
  (11,)

  """

  def _featurize(self, datapoint: str, **kwargs) -> AtomicConformation:
    """Calculate features for a single datapoint.

    Parameters
    ----------
    datapoint: str
      This is expected to be either a filename or a SMILES string.
    """
    from rdkit import Chem
    from rdkit.Chem import AllChem
    if datapoint.endswith('.pdb'):
      mols = [Chem.MolFromPDBFile(datapoint, removeHs=False)]
    elif datapoint.endswith('.sdf'):
      supplier = Chem.SDMolSupplier(datapoint, removeHs=False)
      mols = [mol for mol in supplier]
    elif datapoint.endswith('.mol2'):
      mols = [Chem.MolFromMol2File(datapoint, removeHs=False)]
    else:
      mol = Chem.MolFromSmiles(datapoint)
      # SMILES is unique, so set a canonical order of atoms
      new_order = Chem.rdmolfiles.CanonicalRankAtoms(mol)
      mol = Chem.rdmolops.RenumberAtoms(mol, new_order)
      # Add hydrogens and generate a conformation.
      mol = Chem.AddHs(mol)
      AllChem.EmbedMolecule(mol, AllChem.ETKDGv3())
      mols = [mol]

    # Record properties of the molecules.

    positions = []
    properties = []
    for mol in mols:
      positions.append(mol.GetConformer(0).GetPositions())
      AllChem.ComputeGasteigerCharges(mol)
      for atom in mol.GetAtoms():
        atomic_num = atom.GetAtomicNum()
        formal_charge = atom.GetFormalCharge()
        partial_charge = get_partial_charge(atom)
        properties.append([atomic_num, formal_charge, partial_charge])

    # Create the output object.

    names = ['atomic number', 'formal charge', 'partial charge']
    return AtomicConformation(
        np.concatenate(positions).astype(np.float32),
        np.array(properties, dtype=np.float32), names)



import logging
import numpy as np
from deepchem.utils.typing import RDKitBond, RDKitMol, List
from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.utils.typing import OneOrMany

from typing import Optional

logger = logging.getLogger(__name__)


class GraphMatrix:
  """
  This is class used to store data for MolGAN neural networks.

  Parameters
  ----------
  node_features: np.ndarray
    Node feature matrix with shape [num_nodes, num_node_features]
  edge_features: np.ndarray,
    Edge feature matrix with shape [num_nodes, num_nodes]

  Returns
  -------
  graph: GraphMatrix
    A molecule graph with some features.
  """

  def __init__(self, adjacency_matrix: np.ndarray, node_features: np.ndarray):
    self.adjacency_matrix = adjacency_matrix
    self.node_features = node_features


class MolGanFeaturizer(MolecularFeaturizer):
  """
  Featurizer for MolGAN de-novo molecular generation [1]_.
  The default representation is in form of GraphMatrix object.
  It is wrapper for two matrices containing atom and bond type information.
  The class also provides reverse capabilities.

  Examples
  --------
  >>> import deepchem as dc
  >>> from rdkit import Chem
  >>> rdkit_mol, smiles_mol = Chem.MolFromSmiles('CCC'), 'C1=CC=CC=C1'
  >>> molecules = [rdkit_mol, smiles_mol]
  >>> featurizer = dc.feat.MolGanFeaturizer()
  >>> features = featurizer.featurize(molecules)
  >>> len(features) # 2 molecules
  2
  >>> type(features[0])
  <class 'deepchem.feat.molecule_featurizers.molgan_featurizer.GraphMatrix'>
  >>> molecules = featurizer.defeaturize(features) # defeaturization
  >>> type(molecules[0])
  <class 'rdkit.Chem.rdchem.Mol'>

  """

  def __init__(
      self,
      max_atom_count: int = 9,
      kekulize: bool = True,
      bond_labels: List[RDKitBond] = None,
      atom_labels: List[int] = None,
  ):
    """
    Parameters
    ----------
    max_atom_count: int, default 9
      Maximum number of atoms used for creation of adjacency matrix.
      Molecules cannot have more atoms than this number
      Implicit hydrogens do not count.
    kekulize: bool, default True
      Should molecules be kekulized.
      Solves number of issues with defeaturization when used.
    bond_labels: List[RDKitBond]
      List of types of bond used for generation of adjacency matrix
    atom_labels: List[int]
      List of atomic numbers used for generation of node features

    References
    ---------
    .. [1] Nicola De Cao et al. "MolGAN: An implicit generative model for
       small molecular graphs" (2018), https://arxiv.org/abs/1805.11973
    """

    self.max_atom_count = max_atom_count
    self.kekulize = kekulize

    try:
      from rdkit import Chem
    except ModuleNotFoundError:
      raise ImportError("This class requires RDKit to be installed.")

    # bond labels
    if bond_labels is None:
      self.bond_labels = [
          Chem.rdchem.BondType.ZERO,
          Chem.rdchem.BondType.SINGLE,
          Chem.rdchem.BondType.DOUBLE,
          Chem.rdchem.BondType.TRIPLE,
          Chem.rdchem.BondType.AROMATIC,
      ]
    else:
      self.bond_labels = bond_labels

    # atom labels
    if atom_labels is None:
      self.atom_labels = [0, 6, 7, 8, 9]  # C,N,O,F
    else:
      self.atom_labels = atom_labels

    # create bond encoders and decoders
    self.bond_encoder = {l: i for i, l in enumerate(self.bond_labels)}
    self.bond_decoder = {i: l for i, l in enumerate(self.bond_labels)}
    # create atom encoders and decoders
    self.atom_encoder = {l: i for i, l in enumerate(self.atom_labels)}
    self.atom_decoder = {i: l for i, l in enumerate(self.atom_labels)}

  def _featurize(self, datapoint: RDKitMol, **kwargs) -> Optional[GraphMatrix]:
    """
    Calculate adjacency matrix and nodes features for RDKitMol.
    It strips any chirality and charges

    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit mol object.

    Returns
    -------
    graph: GraphMatrix
      A molecule graph with some features.
    """

    try:
      from rdkit import Chem
    except ModuleNotFoundError:
      raise ImportError("This method requires RDKit to be installed.")
    if 'mol' in kwargs:
      datapoint = kwargs.get("mol")
      raise DeprecationWarning(
          'Mol is being phased out as a parameter, please pass "datapoint" instead.'
      )

    if self.kekulize:
      Chem.Kekulize(datapoint)

    A = np.zeros(
        shape=(self.max_atom_count, self.max_atom_count), dtype=np.float32)
    bonds = datapoint.GetBonds()

    begin, end = [b.GetBeginAtomIdx() for b in bonds], [
        b.GetEndAtomIdx() for b in bonds
    ]
    bond_type = [self.bond_encoder[b.GetBondType()] for b in bonds]

    A[begin, end] = bond_type
    A[end, begin] = bond_type

    degree = np.sum(
        A[:datapoint.GetNumAtoms(), :datapoint.GetNumAtoms()], axis=-1)
    X = np.array(
        [
            self.atom_encoder[atom.GetAtomicNum()]
            for atom in datapoint.GetAtoms()
        ] + [0] * (self.max_atom_count - datapoint.GetNumAtoms()),
        dtype=np.int32,
    )
    graph = GraphMatrix(A, X)

    return graph if (degree > 0).all() else None

  def _defeaturize(self,
                   graph_matrix: GraphMatrix,
                   sanitize: bool = True,
                   cleanup: bool = True) -> RDKitMol:
    """
    Recreate RDKitMol from GraphMatrix object.
    Same featurizer need to be used for featurization and defeaturization.
    It only recreates bond and atom types, any kind of additional features
    like chirality or charge are not included.
    Therefore, any checks of type: original_smiles == defeaturized_smiles
    will fail on chiral or charged compounds.

    Parameters
    ----------
    graph_matrix: GraphMatrix
      GraphMatrix object.
    sanitize: bool, default True
      Should RDKit sanitization be included in the process.
    cleanup: bool, default True
      Splits salts and removes compounds with "*" atom types

    Returns
    -------
    mol: RDKitMol object
      RDKitMol object representing molecule.
    """

    try:
      from rdkit import Chem
    except ModuleNotFoundError:
      raise ImportError("This method requires RDKit to be installed.")

    if not isinstance(graph_matrix, GraphMatrix):
      return None

    node_labels = graph_matrix.node_features
    edge_labels = graph_matrix.adjacency_matrix

    mol = Chem.RWMol()

    for node_label in node_labels:
      mol.AddAtom(Chem.Atom(self.atom_decoder[node_label]))

    for start, end in zip(*np.nonzero(edge_labels)):
      if start > end:
        mol.AddBond(
            int(start), int(end), self.bond_decoder[edge_labels[start, end]])

    if sanitize:
      try:
        Chem.SanitizeMol(mol)
      except Exception:
        mol = None

    if cleanup:
      try:
        smiles = Chem.MolToSmiles(mol)
        smiles = max(smiles.split("."), key=len)
        if "*" not in smiles:
          mol = Chem.MolFromSmiles(smiles)
        else:
          mol = None
      except Exception:
        mol = None

    return mol

  def defeaturize(self, graphs: OneOrMany[GraphMatrix],
                  log_every_n: int = 1000) -> np.ndarray:
    """
    Calculates molecules from corresponding GraphMatrix objects.

    Parameters
    ----------
    graphs: GraphMatrix / iterable
      GraphMatrix object or corresponding iterable
    log_every_n: int, default 1000
      Logging messages reported every `log_every_n` samples.

    Returns
    -------
    features: np.ndarray
      A numpy array containing RDKitMol objext.
    """

    # Special case handling of single molecule
    if isinstance(graphs, GraphMatrix):
      graphs = [graphs]
    else:
      # Convert iterables to list
      graphs = list(graphs)

    molecules = []
    for i, gr in enumerate(graphs):
      if i % log_every_n == 0:
        logger.info("Featurizing datapoint %i" % i)

      try:
        molecules.append(self._defeaturize(gr))
      except Exception as e:
        logger.warning(
            "Failed to defeaturize datapoint %d, %s. Appending empty array",
            i,
            gr,
        )
        logger.warning("Exception message: {}".format(e))
        molecules.append(np.array([]))

    return np.asarray(molecules)



"""Optimizers and related classes for use with TensorGraph."""

import math

from typing import Dict, Union, Optional


class Optimizer(object):
  """An algorithm for optimizing a model.

  This is an abstract class.  Subclasses represent specific optimization algorithms.
  """

  def __init__(self, learning_rate: "Union[float, LearningRateSchedule]"):
    """This constructor should only be called by subclasses.

    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    """
    self.learning_rate = learning_rate

  def _create_tf_optimizer(self, global_step):
    """Construct a TensorFlow optimizer.

    Parameters
    ----------
    global_step: tensor
      a tensor containing the global step index during optimization, used for learning rate decay

    Returns
    -------
    a new TensorFlow optimizer implementing the algorithm
    """
    raise NotImplementedError("Subclasses must implement this")

  def _create_pytorch_optimizer(self, params):
    """Construct a PyTorch optimizer.

    Parameters
    ----------
    params: Iterable
      the model parameters to optimize

    Returns
    -------
    a new PyTorch optimizer implementing the algorithm
    """
    raise NotImplementedError("Subclasses must implement this")

  def _create_jax_optimizer(self):
    """Construct a Jax optimizer.

    Returns
    -------
    a new Optax optimizer optax.GradientTransformation implementing the algorithm
    """
    raise NotImplementedError("Subclasses must implement this")


class LearningRateSchedule(object):
  """A schedule for changing the learning rate over the course of optimization.

  This is an abstract class.  Subclasses represent specific schedules.
  """

  def _create_tf_tensor(self, global_step):
    """Construct a tensor that equals the learning rate.

    Parameters
    ----------
    global_step: tensor
      a tensor containing the global step index during optimization

    Returns
    -------
    a tensor that equals the learning rate
    """
    raise NotImplementedError("Subclasses must implement this")

  def _create_pytorch_schedule(self, optimizer):
    """Construct a PyTorch learning rate scheduler.

    Parameters
    ----------
    optimizer: torch.optim.Optimizer
      the Optimizer whose learning rate will be modified

    Returns
    -------
    a PyTorch scheduler implementing the schedule
    """
    raise NotImplementedError("Subclasses must implement this")

  def _create_jax_schedule(self, learning_rate):
    """Construct a Jax learning rate scheduler using optax.

    Parameters
    ----------
    learning_rate: float
      the initial learning rate that will be modified

    Returns
    -------
    a optax scheduler implementing the schedule
    """
    raise NotImplementedError("Subclasses must implement this")


class AdaGrad(Optimizer):
  """The AdaGrad optimization algorithm.

  Adagrad is an optimizer with parameter-specific learning rates, which are
  adapted relative to how frequently a parameter gets updated during training.
  The more updates a parameter receives, the smaller the updates. See [1]_ for
  a full reference for the algorithm.

  References
  ----------
  .. [1] Duchi, John, Elad Hazan, and Yoram Singer. "Adaptive subgradient
     methods for online learning and stochastic optimization." Journal of machine
     learning research 12.7 (2011).
  """

  def __init__(self,
               learning_rate: Union[float, LearningRateSchedule] = 0.001,
               initial_accumulator_value: float = 0.1,
               epsilon: float = 1e-07):
    """Construct an AdaGrad optimizer.
    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    initial_accumulator_value: float
      a parameter of the AdaGrad algorithm
    epsilon: float
      a parameter of the AdaGrad algorithm

    """
    super(AdaGrad, self).__init__(learning_rate)
    self.initial_accumulator_value = initial_accumulator_value
    self.epsilon = epsilon

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tf.keras.optimizers.Adagrad(
        learning_rate=learning_rate,
        initial_accumulator_value=self.initial_accumulator_value,
        epsilon=self.epsilon)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.Adagrad(
        params,
        lr,
        initial_accumulator_value=self.initial_accumulator_value,
        eps=self.epsilon)

  def _create_jax_optimizer(self):
    import optax
    process = []
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
      last_process = optax.scale(-1.0)
    else:
      lr = self.learning_rate
      last_process = optax.scale(-1.0 * lr)

    process.append(
        optax.scale_by_rss(
            initial_accumulator_value=self.initial_accumulator_value,
            eps=self.epsilon))
    process.append(last_process)
    return optax.chain(*process)


class Adam(Optimizer):
  """The Adam optimization algorithm."""

  def __init__(self,
               learning_rate: Union[float, LearningRateSchedule] = 0.001,
               beta1: float = 0.9,
               beta2: float = 0.999,
               epsilon: float = 1e-08):
    """Construct an Adam optimizer.

    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    beta1: float
      a parameter of the Adam algorithm
    beta2: float
      a parameter of the Adam algorithm
    epsilon: float
      a parameter of the Adam algorithm
    """
    super(Adam, self).__init__(learning_rate)
    self.beta1 = beta1
    self.beta2 = beta2
    self.epsilon = epsilon

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tf.keras.optimizers.Adam(
        learning_rate=learning_rate,
        beta_1=self.beta1,
        beta_2=self.beta2,
        epsilon=self.epsilon)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.Adam(params, lr, (self.beta1, self.beta2), self.epsilon)

  def _create_jax_optimizer(self):
    import optax
    process = []
    if isinstance(self.learning_rate, LearningRateSchedule):
      scheduler = self.learning_rate._create_jax_schedule()
      process.append(optax.scale_by_schedule(scheduler))
      last_process = optax.scale(-1.0)
    else:
      lr = self.learning_rate
      last_process = optax.scale(-1.0 * lr)

    process.append(
        optax.scale_by_adam(b1=self.beta1, b2=self.beta2, eps=self.epsilon))
    process.append(last_process)
    return optax.chain(*process)


class SparseAdam(Optimizer):
  """The Sparse Adam optimization algorithm, also known as Lazy Adam.
  Sparse Adam is suitable for sparse tensors. It handles sparse updates more efficiently.
  It only updates moving-average accumulators for sparse variable indices that appear in the current batch, rather than updating the accumulators for all indices.
  """

  def __init__(self,
               learning_rate: Union[float, LearningRateSchedule] = 0.001,
               beta1: float = 0.9,
               beta2: float = 0.999,
               epsilon: float = 1e-08):
    """Construct an Adam optimizer.

    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    beta1: float
      a parameter of the SparseAdam algorithm
    beta2: float
      a parameter of the SparseAdam algorithm
    epsilon: float
      a parameter of the SparseAdam algorithm
    """
    super(SparseAdam, self).__init__(learning_rate)
    self.beta1 = beta1
    self.beta2 = beta2
    self.epsilon = epsilon

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    import tensorflow_addons as tfa
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tfa.optimizers.LazyAdam(
        learning_rate=learning_rate,
        beta_1=self.beta1,
        beta_2=self.beta2,
        epsilon=self.epsilon)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.SparseAdam(params, lr, (self.beta1, self.beta2),
                                  self.epsilon)


class AdamW(Optimizer):
  """The AdamW optimization algorithm.
  AdamW is a variant of Adam, with improved weight decay.
  In Adam, weight decay is implemented as: weight_decay (float, optional) – weight decay (L2 penalty) (default: 0)
  In AdamW, weight decay is implemented as: weight_decay (float, optional) – weight decay coefficient (default: 1e-2)
  """

  def __init__(self,
               learning_rate: Union[float, LearningRateSchedule] = 0.001,
               weight_decay: Union[float, LearningRateSchedule] = 0.01,
               beta1: float = 0.9,
               beta2: float = 0.999,
               epsilon: float = 1e-08,
               amsgrad: bool = False):
    """Construct an AdamW optimizer.
    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    weight_decay: float or LearningRateSchedule
      weight decay coefficient for AdamW
    beta1: float
      a parameter of the Adam algorithm
    beta2: float
      a parameter of the Adam algorithm
    epsilon: float
      a parameter of the Adam algorithm
    amsgrad: bool
      If True, will use the AMSGrad variant of AdamW (from "On the Convergence of Adam and Beyond"), else will use the original algorithm.
    """
    super(AdamW, self).__init__(learning_rate)
    self.weight_decay = weight_decay
    self.beta1 = beta1
    self.beta2 = beta2
    self.epsilon = epsilon
    self.amsgrad = amsgrad

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    import tensorflow_addons as tfa
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tfa.optimizers.AdamW(
        weight_decay=self.weight_decay,
        learning_rate=learning_rate,
        beta_1=self.beta1,
        beta_2=self.beta2,
        epsilon=self.epsilon,
        amsgrad=self.amsgrad)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.AdamW(params, lr, (self.beta1, self.beta2), self.epsilon,
                             self.weight_decay, self.amsgrad)

  def _create_jax_optimizer(self):
    import optax
    process = []
    if isinstance(self.learning_rate, LearningRateSchedule):
      scheduler = self.learning_rate._create_jax_schedule()
      process.append(optax.scale_by_schedule(scheduler))
      last_process = optax.scale(-1.0)
    else:
      lr = self.learning_rate
      last_process = optax.scale(-1.0 * lr)

    process.append(
        optax.scale_by_adam(
            b1=self.beta1, b2=self.beta2, eps=self.epsilon, eps_root=0.0))
    process.append(optax.add_decayed_weights(self.weight_decay, None))
    process.append(last_process)
    return optax.chain(*process)


class RMSProp(Optimizer):
  """RMSProp Optimization algorithm."""

  def __init__(self,
               learning_rate: Union[float, LearningRateSchedule] = 0.001,
               momentum: float = 0.0,
               decay: float = 0.9,
               epsilon: float = 1e-10):
    """Construct an RMSProp Optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning_rate used for optimization
        momentum: float, default 0.0
            a parameter of the RMSProp algorithm
        decay: float, default 0.9
            a parameter of the RMSProp algorithm
        epsilon: float, default 1e-10
            a parameter of the RMSProp algorithm
        """
    super(RMSProp, self).__init__(learning_rate)
    self.momentum = momentum
    self.decay = decay
    self.epsilon = epsilon

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tf.keras.optimizers.RMSprop(
        learning_rate=learning_rate,
        momentum=self.momentum,
        rho=self.decay,
        epsilon=self.epsilon)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.RMSprop(
        params, lr, alpha=self.decay, eps=self.epsilon, momentum=self.momentum)

  def _create_jax_optimizer(self):
    import optax
    process = []
    if isinstance(self.learning_rate, LearningRateSchedule):
      scheduler = self.learning_rate._create_jax_schedule()
      process.append(optax.scale_by_schedule(scheduler))
      last_process = optax.scale(-1.0)
    else:
      lr = self.learning_rate
      last_process = optax.scale(-1.0 * lr)

    process.append(
        optax.scale_by_rms(
            decay=self.decay, eps=self.epsilon, initial_scale=0.0))
    if self.momentum is not None or self.momentum != 0.0:
      process.append(optax.trace(decay=self.momentum, nesterov=False))
    process.append(last_process)
    return optax.chain(*process)


class GradientDescent(Optimizer):
  """The gradient descent optimization algorithm."""

  def __init__(self, learning_rate: Union[float, LearningRateSchedule] = 0.001):
    """Construct a gradient descent optimizer.

    Parameters
    ----------
    learning_rate: float or LearningRateSchedule
      the learning rate to use for optimization
    """
    super(GradientDescent, self).__init__(learning_rate)

  def _create_tf_optimizer(self, global_step):
    import tensorflow as tf
    if isinstance(self.learning_rate, LearningRateSchedule):
      learning_rate = self.learning_rate._create_tf_tensor(global_step)
    else:
      learning_rate = self.learning_rate
    return tf.keras.optimizers.SGD(learning_rate=learning_rate)

  def _create_pytorch_optimizer(self, params):
    import torch
    if isinstance(self.learning_rate, LearningRateSchedule):
      lr = self.learning_rate.initial_rate
    else:
      lr = self.learning_rate
    return torch.optim.SGD(params, lr)

  def _create_jax_optimizer(self):
    import optax
    process = []
    if isinstance(self.learning_rate, LearningRateSchedule):
      scheduler = self.learning_rate._create_jax_schedule()
      process.append(optax.scale_by_schedule(scheduler))
      last_process = optax.scale(-1.0)
    else:
      lr = self.learning_rate
      last_process = optax.scale(-1.0 * lr)
    process.append(last_process)
    return optax.chain(*process)


class ExponentialDecay(LearningRateSchedule):
  """A learning rate that decreases exponentially with the number of training steps."""

  def __init__(self,
               initial_rate: float,
               decay_rate: float,
               decay_steps: int,
               staircase: bool = True):
    """Create an exponentially decaying learning rate.

    The learning rate starts as initial_rate.  Every decay_steps training steps, it is multiplied by decay_rate.

    Parameters
    ----------
    initial_rate: float
      the initial learning rate
    decay_rate: float
      the base of the exponential
    decay_steps: int
      the number of training steps over which the rate decreases by decay_rate
    staircase: bool
      if True, the learning rate decreases by discrete jumps every decay_steps.
      if False, the learning rate decreases smoothly every step
    """
    self.initial_rate = initial_rate
    self.decay_rate = decay_rate
    self.decay_steps = decay_steps
    self.staircase = staircase

  def _create_tf_tensor(self, global_step):
    import tensorflow as tf
    return tf.keras.optimizers.schedules.ExponentialDecay(
        initial_learning_rate=self.initial_rate,
        decay_rate=self.decay_rate,
        decay_steps=self.decay_steps,
        staircase=self.staircase)(global_step)

  def _create_pytorch_schedule(self, optimizer):
    import torch
    if self.staircase:
      return torch.optim.lr_scheduler.StepLR(optimizer, self.decay_steps,
                                             self.decay_rate)
    return torch.optim.lr_scheduler.ExponentialLR(
        optimizer, math.pow(self.decay_rate, 1 / self.decay_steps))

  def _create_jax_schedule(self):
    import optax
    return optax.exponential_decay(
        init_value=self.initial_rate,
        transition_steps=self.decay_steps,
        decay_rate=self.decay_rate,
        staircase=self.staircase)


class PolynomialDecay(LearningRateSchedule):
  """A learning rate that decreases from an initial value to a final value over a fixed number of training steps."""

  def __init__(self,
               initial_rate: float,
               final_rate: float,
               decay_steps: int,
               power: float = 1.0):
    """Create a smoothly decaying learning rate.

    The learning rate starts as initial_rate.  It smoothly decreases to final_rate over decay_steps training steps.
    It decays as a function of (1-step/decay_steps)**power.  Once the final rate is reached, it remains there for
    the rest of optimization.

    Parameters
    ----------
    initial_rate: float
      the initial learning rate
    final_rate: float
      the final learning rate
    decay_steps: int
      the number of training steps over which the rate decreases from initial_rate to final_rate
    power: float
      the exponent controlling the shape of the decay
    """
    self.initial_rate = initial_rate
    self.final_rate = final_rate
    self.decay_steps = decay_steps
    self.power = power

  def _create_tf_tensor(self, global_step):
    import tensorflow as tf
    return tf.keras.optimizers.schedules.PolynomialDecay(
        initial_learning_rate=self.initial_rate,
        end_learning_rate=self.final_rate,
        decay_steps=self.decay_steps,
        power=self.power)(global_step)

  def _create_pytorch_schedule(self, optimizer):

    def f(step):
      t = min(step, self.decay_steps) / self.decay_steps
      return ((self.initial_rate - self.final_rate) *
              (1 - t)**self.power) + self.final_rate

    import torch
    return torch.optim.lr_scheduler.LambdaLR(optimizer, f)

  def _create_jax_schedule(self):
    import optax
    return optax.polynomial_schedule(
        init_value=self.initial_rate,
        end_value=self.final_rate,
        power=self.power,
        transition_steps=self.decay_steps)


class LinearCosineDecay(LearningRateSchedule):
  """Applies linear cosine decay to the learning rate"""

  def __init__(self,
               initial_rate: float,
               decay_steps: int,
               alpha: float = 0.0,
               beta: float = 0.001,
               num_periods: float = 0.5):
    """
    Parameters
    ----------
    learning_rate : float
    initial learning rate
    decay_steps : int
    number of steps to decay over
    num_periods : number of periods in the cosine part of the decay
    """

    self.initial_rate = initial_rate
    self.decay_steps = decay_steps
    self.alpha = alpha
    self.beta = beta
    self.num_periods = num_periods

  def _create_tf_tensor(self, global_step):
    import tensorflow as tf
    return tf.compat.v1.train.linear_cosine_decay(
        learning_rate=self.initial_rate,
        global_step=global_step,
        decay_steps=self.decay_steps,
        alpha=self.alpha,
        beta=self.beta,
        num_periods=self.num_periods)

  def _create_pytorch_schedule(self, optimizer):

    def f(step):
      t = min(step, self.decay_steps) / self.decay_steps
      linear_decay = 1 - t
      cosine_decay = 0.5 * (1 + math.cos(math.pi * 2 * self.num_periods * t))
      decayed = (self.alpha + linear_decay) * cosine_decay + self.beta
      return self.initial_rate * decayed

    import torch
    return torch.optim.lr_scheduler.LambdaLR(optimizer, f)

  def _create_jax_schedule(self):
    import optax
    return optax.cosine_decay_schedule(
        init_value=self.initial_rate,
        decay_steps=self.decay_steps,
        alpha=self.alpha)


class PiecewiseConstantSchedule(LearningRateSchedule):
  """Applies scheduler which multiplies by a constant factor on the boundaries"""

  def __init__(self,
               initial_rate: float,
               boundaries_and_scales: Optional[Dict[int, float]] = None):
    """
    Parameters
    ----------
    init_value : float
      initial learning rate
    boundaries_and_scales:
      A map from boundaries b_i to non-negative scaling factors f_i. For any step
      count s, the schedule returns init_v scaled by the product of all factors f_i
      such that b_i < s.
    """
    self.initial_rate = initial_rate
    self.boundaries_and_scales = boundaries_and_scales

  def _create_jax_schedule(self):
    import optax
    return optax.piecewise_constant_schedule(
        init_value=self.initial_rate,
        boundaries_and_scales=self.boundaries_and_scales)



"""
DGL-based AttentiveFP for graph property prediction.
"""
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel


class AttentiveFP(nn.Module):
  """Model for Graph Property Prediction.

  This model proceeds as follows:

  * Combine node features and edge features for initializing node representations,
    which involves a round of message passing
  * Update node representations with multiple rounds of message passing
  * For each graph, compute its representation by combining the representations
    of all nodes in it, which involves a gated recurrent unit (GRU).
  * Perform the final prediction using a linear layer

  Examples
  --------

  >>> import deepchem as dc
  >>> import dgl
  >>> from deepchem.models import AttentiveFP
  >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
  >>> featurizer = dc.feat.MolGraphConvFeaturizer(use_edges=True)
  >>> graphs = featurizer.featurize(smiles)
  >>> print(type(graphs[0]))
  <class 'deepchem.feat.graph_data.GraphData'>
  >>> dgl_graphs = [graphs[i].to_dgl_graph(self_loop=True) for i in range(len(graphs))]
  >>> # Batch two graphs into a graph of two connected components
  >>> batch_dgl_graph = dgl.batch(dgl_graphs)
  >>> model = AttentiveFP(n_tasks=1, mode='regression')
  >>> preds = model(batch_dgl_graph)
  >>> print(type(preds))
  <class 'torch.Tensor'>
  >>> preds.shape == (2, 1)
  True

  References
  ----------
  .. [1] Zhaoping Xiong, Dingyan Wang, Xiaohong Liu, Feisheng Zhong, Xiaozhe Wan, Xutong Li,
         Zhaojun Li, Xiaomin Luo, Kaixian Chen, Hualiang Jiang, and Mingyue Zheng. "Pushing
         the Boundaries of Molecular Representation for Drug Discovery with the Graph Attention
         Mechanism." Journal of Medicinal Chemistry. 2020, 63, 16, 8749–8760.

  Notes
  -----
  This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
  (https://github.com/awslabs/dgl-lifesci) to be installed.
  """

  def __init__(self,
               n_tasks: int,
               num_layers: int = 2,
               num_timesteps: int = 2,
               graph_feat_size: int = 200,
               dropout: float = 0.,
               mode: str = 'regression',
               number_atom_features: int = 30,
               number_bond_features: int = 11,
               n_classes: int = 2,
               nfeat_name: str = 'x',
               efeat_name: str = 'edge_attr'):
    """
    Parameters
    ----------
    n_tasks: int
      Number of tasks.
    num_layers: int
      Number of graph neural network layers, i.e. number of rounds of message passing.
      Default to 2.
    num_timesteps: int
      Number of time steps for updating graph representations with a GRU. Default to 2.
    graph_feat_size: int
      Size for graph representations. Default to 200.
    dropout: float
      Dropout probability. Default to 0.
    mode: str
      The model type, 'classification' or 'regression'. Default to 'regression'.
    number_atom_features: int
      The length of the initial atom feature vectors. Default to 30.
    number_bond_features: int
      The length of the initial bond feature vectors. Default to 11.
    n_classes: int
      The number of classes to predict per task
      (only used when ``mode`` is 'classification'). Default to 2.
    nfeat_name: str
      For an input graph ``g``, the model assumes that it stores node features in
      ``g.ndata[nfeat_name]`` and will retrieve input node features from that.
      Default to 'x'.
    efeat_name: str
      For an input graph ``g``, the model assumes that it stores edge features in
      ``g.edata[efeat_name]`` and will retrieve input edge features from that.
      Default to 'edge_attr'.
    """
    try:
      import dgl
    except:
      raise ImportError('This class requires dgl.')
    try:
      import dgllife
    except:
      raise ImportError('This class requires dgllife.')

    if mode not in ['classification', 'regression']:
      raise ValueError("mode must be either 'classification' or 'regression'")

    super(AttentiveFP, self).__init__()

    self.n_tasks = n_tasks
    self.mode = mode
    self.n_classes = n_classes
    self.nfeat_name = nfeat_name
    self.efeat_name = efeat_name
    if mode == 'classification':
      out_size = n_tasks * n_classes
    else:
      out_size = n_tasks

    from dgllife.model import AttentiveFPPredictor as DGLAttentiveFPPredictor

    self.model = DGLAttentiveFPPredictor(
        node_feat_size=number_atom_features,
        edge_feat_size=number_bond_features,
        num_layers=num_layers,
        num_timesteps=num_timesteps,
        graph_feat_size=graph_feat_size,
        n_tasks=out_size,
        dropout=dropout)

  def forward(self, g):
    """Predict graph labels

    Parameters
    ----------
    g: DGLGraph
      A DGLGraph for a batch of graphs. It stores the node features in
      ``dgl_graph.ndata[self.nfeat_name]`` and edge features in
      ``dgl_graph.edata[self.efeat_name]``.

    Returns
    -------
    torch.Tensor
      The model output.

      * When self.mode = 'regression',
        its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
      * When self.mode = 'classification', the output consists of probabilities
        for classes. Its shape will be
        ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)`` if self.n_tasks > 1;
        its shape will be ``(dgl_graph.batch_size, self.n_classes)`` if self.n_tasks is 1.
    torch.Tensor, optional
      This is only returned when self.mode = 'classification', the output consists of the
      logits for classes before softmax.
    """
    node_feats = g.ndata[self.nfeat_name]
    edge_feats = g.edata[self.efeat_name]
    out = self.model(g, node_feats, edge_feats)

    if self.mode == 'classification':
      if self.n_tasks == 1:
        logits = out.view(-1, self.n_classes)
        softmax_dim = 1
      else:
        logits = out.view(-1, self.n_tasks, self.n_classes)
        softmax_dim = 2
      proba = F.softmax(logits, dim=softmax_dim)
      return proba, logits
    else:
      return out


class AttentiveFPModel(TorchModel):
  """Model for Graph Property Prediction.

  This model proceeds as follows:

  * Combine node features and edge features for initializing node representations,
    which involves a round of message passing
  * Update node representations with multiple rounds of message passing
  * For each graph, compute its representation by combining the representations
    of all nodes in it, which involves a gated recurrent unit (GRU).
  * Perform the final prediction using a linear layer

  Examples
  --------
  >>> import deepchem as dc
  >>> from deepchem.models import AttentiveFPModel
  >>> # preparing dataset
  >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
  >>> labels = [0., 1.]
  >>> featurizer = dc.feat.MolGraphConvFeaturizer(use_edges=True)
  >>> X = featurizer.featurize(smiles)
  >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
  >>> # training model
  >>> model = AttentiveFPModel(mode='classification', n_tasks=1,
  ...    batch_size=16, learning_rate=0.001)
  >>> loss = model.fit(dataset, nb_epoch=5)

  References
  ----------
  .. [1] Zhaoping Xiong, Dingyan Wang, Xiaohong Liu, Feisheng Zhong, Xiaozhe Wan, Xutong Li,
         Zhaojun Li, Xiaomin Luo, Kaixian Chen, Hualiang Jiang, and Mingyue Zheng. "Pushing
         the Boundaries of Molecular Representation for Drug Discovery with the Graph
         Attention Mechanism." Journal of Medicinal Chemistry. 2020, 63, 16, 8749–8760.

  Notes
  -----
  This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
  (https://github.com/awslabs/dgl-lifesci) to be installed.
  """

  def __init__(self,
               n_tasks: int,
               num_layers: int = 2,
               num_timesteps: int = 2,
               graph_feat_size: int = 200,
               dropout: float = 0.,
               mode: str = 'regression',
               number_atom_features: int = 30,
               number_bond_features: int = 11,
               n_classes: int = 2,
               self_loop: bool = True,
               **kwargs):
    """
    Parameters
    ----------
    n_tasks: int
      Number of tasks.
    num_layers: int
      Number of graph neural network layers, i.e. number of rounds of message passing.
      Default to 2.
    num_timesteps: int
      Number of time steps for updating graph representations with a GRU. Default to 2.
    graph_feat_size: int
      Size for graph representations. Default to 200.
    dropout: float
      Dropout probability. Default to 0.
    mode: str
      The model type, 'classification' or 'regression'. Default to 'regression'.
    number_atom_features: int
      The length of the initial atom feature vectors. Default to 30.
    number_bond_features: int
      The length of the initial bond feature vectors. Default to 11.
    n_classes: int
      The number of classes to predict per task
      (only used when ``mode`` is 'classification'). Default to 2.
    self_loop: bool
      Whether to add self loops for the nodes, i.e. edges from nodes to themselves.
      When input graphs have isolated nodes, self loops allow preserving the original feature
      of them in message passing. Default to True.
    kwargs
      This can include any keyword argument of TorchModel.
    """
    model = AttentiveFP(
        n_tasks=n_tasks,
        num_layers=num_layers,
        num_timesteps=num_timesteps,
        graph_feat_size=graph_feat_size,
        dropout=dropout,
        mode=mode,
        number_atom_features=number_atom_features,
        number_bond_features=number_bond_features,
        n_classes=n_classes)
    if mode == 'regression':
      loss: Loss = L2Loss()
      output_types = ['prediction']
    else:
      loss = SparseSoftmaxCrossEntropy()
      output_types = ['prediction', 'loss']
    super(AttentiveFPModel, self).__init__(
        model, loss=loss, output_types=output_types, **kwargs)

    self._self_loop = self_loop

  def _prepare_batch(self, batch):
    """Create batch data for AttentiveFP.

    Parameters
    ----------
    batch: tuple
      The tuple is ``(inputs, labels, weights)``.

    Returns
    -------
    inputs: DGLGraph
      DGLGraph for a batch of graphs.
    labels: list of torch.Tensor or None
      The graph labels.
    weights: list of torch.Tensor or None
      The weights for each sample or sample/task pair converted to torch.Tensor.
    """
    try:
      import dgl
    except:
      raise ImportError('This class requires dgl.')

    inputs, labels, weights = batch
    dgl_graphs = [
        graph.to_dgl_graph(self_loop=self._self_loop) for graph in inputs[0]
    ]
    inputs = dgl.batch(dgl_graphs).to(self.device)
    _, labels, weights = super(AttentiveFPModel, self)._prepare_batch(
        ([], labels, weights))
    return inputs, labels, weights
